@@ -1,14 +1,14 @@
 from dataclasses import dataclass
 from datetime import datetime
-from typing import Iterable, List, Optional
+from typing import List
 
 from loguru import logger
 from pymodbus.constants import Defaults
 
 from mate3.devices import DeviceValues
-from mate3.field_values import FieldRead, FieldValue
 from mate3.modbus_client import CachingModbusClient, ModbusTcpClient, NonCachingModbusClient
-from mate3.sunspec.fields import Field, IntegerField, Mode, Uint16Field, Uint32Field
+from mate3.read import AllModelReads, ModelRead
+from mate3.sunspec.fields import Field, Mode, Uint16Field, Uint32Field
 from mate3.sunspec.models import MODEL_DEVICE_IDS, SunSpecEndModel, SunSpecHeaderModel
 
 
@@ -39,11 +39,19 @@ class Mate3Client:
 
     sunspec_register: int = 40000
 
-    def __init__(self, host: str, port: int = Defaults.Port, cache_path: str = None, cache_only: bool = False):
+    def __init__(
+        self,
+        host: str,
+        port: int = Defaults.Port,
+        cache_path: str = None,
+        cache_only: bool = False,
+        cache_writeable: bool = False,
+    ):
         self.host: str = host
         self.port: int = port
         self._cache_path: str = cache_path
         self._cache_only: bool = cache_only
+        self._cache_writeable: bool = cache_writeable
         self._client: ModbusTcpClient = None
         self._devices: DeviceValues = None
 
@@ -53,11 +61,19 @@ class Mate3Client:
         """
         if self._cache_path is not None:
             self._client = CachingModbusClient(
-                host=self.host, port=self.port, cache_path=self._cache_path, cache_only=self._cache_only
+                host=self.host,
+                port=self.port,
+                cache_path=self._cache_path,
+                cache_only=self._cache_only,
+                writeable=self._cache_writeable,
             )
         else:
             self._client = NonCachingModbusClient(self.host, self.port)
 
+        # Now read everything. Why? Because most use of the API assumes fields have already been read (e.g. to get
+        # the devices, or the addresses of fields, etc.)
+        self.read_all()
+
     def close(self):
         """
         Close the modbus connection to the mate.
@@ -90,21 +106,25 @@ class Mate3Client:
         ranges = []
         previous_range = None
         for field in sorted(fields, key=lambda x: x.start):
-            if previous_range is None or previous_range.end != field.start or previous_range.size >= max_range_size:
+            if (
+                previous_range is None
+                or previous_range.end != field.start
+                or previous_range.size + field.size >= max_range_size
+            ):
                 previous_range = ReadingRange(fields=[field], start=field.start, size=field.size)
                 ranges.append(previous_range)
             else:
                 previous_range.extend(field)
         return ranges
 
-    def _read_model(self, address: int, first: bool, only: Optional[List[Field]] = None):
+    def _read_model(self, device_address: int, first: bool, all_reads: AllModelReads):
         """
         Read an individual model at `address`. Use `first` to specify that this is the first block - see comment below.
         By default reads everything in the model - use `only` to specify a list of Fields to read, if you want to limit.
         """
 
         # Read the first register for the device ID:
-        registers = self._client.read_holding_registers(address=address, count=2)
+        registers = self._client.read_holding_registers(address=device_address, count=2)
 
         # If first, then this is the SunSpecHeaderModel, which has a device ID of Uint32 type not Uint16 like the rest.
         if first:
@@ -126,69 +146,52 @@ class Mate3Client:
 
         # Get the readable fields:
         fields = [field for field in model.__model_fields__ if field.mode in (Mode.R, Mode.RW)]
-        if only is not None:
-            # Make sure we include the length and port number, as we need that elsewhere:
-            fields = [field for field in fields if field in only or field.name in ("length", "port_number")]
 
         # Order fields by start registry, as this is the order in which we will receive the values
         fields = sorted(fields, key=lambda f: f.start)
 
         # Get registers in large ranges, as this drastically improves performance and isn't so demanding of the mate3
-        field_reads = {}
+        model_reads = ModelRead()
         for reading_range in self._get_reading_ranges(fields):
             logger.debug(
                 f"Reading range {reading_range.start} -> {reading_range.end}, of {len(reading_range.fields)} fields"
             )
-            register_number = address + reading_range.start - 1
+            register_number = device_address + reading_range.start - 1  # -1 as starts are 1-indexed in fields
             registers = self._client.read_holding_registers(address=register_number, count=reading_range.size)
             read_time = datetime.now()
 
+            offset = 0
             for field in reading_range.fields:
+                address = register_number + offset
                 try:
-                    implemented, value = field.from_registers(registers[: field.size])
-                    field_reads[field.name] = FieldRead(
-                        value=value, implemented=implemented, time=read_time, scale_factor=None
+                    field_registers = registers[offset : offset + field.size]
+                    if len(field_registers) != field.size:
+                        raise RuntimeError(
+                            "Didn't get the right number of registers from reading range for this field."
+                        )
+                    implemented, raw_value = field.from_registers(field_registers)
+                    model_reads.add(
+                        field.name, raw_value=raw_value, implemented=implemented, address=address, time=read_time
                     )
                 except Exception as e:
                     logger.warning(f"Error reading field {field.name} - so setting as not implemented. Message: {e}")
-                    field_reads[field.name] = FieldRead(
-                        value=None, implemented=False, time=read_time, scale_factor=None
-                    )
-                registers = registers[field.size :]
+                    model_reads.add(field.name, raw_value=None, implemented=False, address=address, time=read_time)
+                offset += field.size
 
-        for field in fields:
-            if isinstance(field, IntegerField) and field.scale_factor is not None:
-                scale_factor = field_reads[field.scale_factor.name]
-                if not scale_factor.implemented:
-                    raise RuntimeError("Scale factor isn't implemented!")
-                field_reads[field.name].scale_factor = scale_factor
+        return model, model_reads
 
-        return model, field_reads
-
-    def read(self, only: Optional[Iterable[FieldValue]] = None):
+    def read_all(self):
         """
-        Read values from the mate. By default (`only=None`), everything is read, but if you want to read only specified
-        fields (which puts less strain on the mate), list the in `only`.
+        Read all values from all devices. If you want to read only specified fields use e.g.
+            client.devices.mate3.system_name.read()
+        This method, however, is optimised for reading everything.
         """
         register = self.sunspec_register
         max_models = 30
         first = True
-        model_field_reads = {}
-        model_addresses = {}
-        if only is not None:
-            only_fields = []
-            for field in only:
-                field = field.field
-                if field.mode not in (Mode.R, Mode.RW):
-                    raise RuntimeError("Can't read from a read-only field!")
-                only_fields.append(field)
-                # Read the scale factor too, if needed:
-                if isinstance(field, IntegerField) and field.scale_factor is not None:
-                    only_fields.append(field.scale_factor)
-        else:
-            only_fields = None
+        all_reads = AllModelReads()
         for _ in range(max_models):
-            model, field_reads = self._read_model(register, first, only_fields)
+            model, model_reads = self._read_model(register, first, all_reads)
             first = False
 
             # Unknown device
@@ -199,55 +202,21 @@ class Mate3Client:
             if model == SunSpecEndModel:
                 break
 
-            # record for use out of loop:
-            model_field_reads.setdefault(model, [])
-            model_field_reads[model].append(field_reads)
-            model_addresses.setdefault(model, [])
-            model_addresses[model].append(register)
+            # Save the model reads for this model:
+            all_reads.add(model, model_reads)
 
             # Move register to next block - that is, add the length of the block, which is what follows after the length
             # field. For normal fields, we'll already have read 2 registers (DI and length) so we must add this to our
             # total increment. For the SunSpecHeaderModel, we need to add 4 as DID is a UInt32 in this case (i.e. 2
             # registers) and there's a model ID field (1 register) and length (1 register)
-            register += field_reads["length"].value + (4 if model == SunSpecHeaderModel else 2)
+            register += model_reads["length"].raw_value + (4 if model == SunSpecHeaderModel else 2)
 
         # create devices if needed:
         if self._devices is None:
-            self._devices = DeviceValues()
+            self._devices = DeviceValues(client=self)
 
         # update:
-        self._devices.update(model_field_reads, model_addresses)
-
-    def write(self):
-        """
-        After having updated values on self.devices, call this function to write them to the mate itself.
-
-        Warning:
-
-            Ensure you have read the LICENSE file before using this feature. Note the
-            section which begins:
-
-                THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-                EXPRESS OR IMPLIED
-
-            Specifically, it is quite possible that you can cause damage to your equipment
-            through use of this feature. Be careful!
-        """
-
-        # TODO: write in ranges?
-        for device in self._devices.connected_devices:
-            for field_value in device.fields():
-                if field_value.dirty:
-                    if field_value.field.mode not in (Mode.RW, Mode.W):
-                        raise RuntimeError(f"Can't write to field {field_value.name}")
-                    logger.info(f"writing {field_value._value_to_write} to {field_value.name}")
-                    address = device.address + field_value.field.start - 1  # -1 since start is 1-indexed
-                    # TODO: should we check that the device at the given address still has the right port?
-                    registers = field_value.field.to_registers(field_value._value_to_write)
-                    logger.debug(f"Setting register {address} to value {registers}")
-
-                    # Do the write
-                    self._client.write_registers(address, registers)
+        self._devices.update(all_reads)
 
     def read_all_modbus_values_unparsed(self):
         """
