@@ -47,7 +47,7 @@ def read(client, args):
             for value in device.fields([Mode.R, Mode.RW]):
                 values[value.field.name] = {
                     "implemented": value.implemented,
-                    "scale_factor": value.scale_factor,
+                    "scale_factor": value.scale_factor.value if value.scale_factor is not None else None,
                     "raw_value": value.raw_value
                     if value.raw_value is None or isinstance(value.raw_value, (str, int, float))
                     else repr(value.raw_value),
